type voeu = {
  rang: Z.t;
  boursier: bool;
  resident: bool;
  }

let ordre_appel (boursiers: voeu Queue.t) (non_boursiers: voeu Queue.t)
                (taux: Z.t) : voeu Queue.t =
  let nbBoursiersTotal = Z.of_int (Queue.length boursiers) in
  let nbVoeux =
    Z.add (Z.of_int (Queue.length boursiers)) (Z.of_int (Queue.length non_boursiers)) in
  let nbAppeles = ref Z.zero in
  let nbBoursiersAppeles = ref Z.zero in
  let ordreAppel = Queue.create () in
  while Z.lt (!nbAppeles) nbVoeux do
    let resteBoursiers = Z.lt (!nbBoursiersAppeles) nbBoursiersTotal in
    let contrainteTauxBoursier =
      resteBoursiers && Z.lt (Z.mul (!nbBoursiersAppeles) (Z.of_string "100")) 
                        (Z.mul taux (Z.add Z.one (!nbAppeles))) in
    if contrainteTauxBoursier
    then
      let v = Queue.pop boursiers in
      nbBoursiersAppeles := Z.succ !nbBoursiersAppeles;
      nbAppeles := Z.succ !nbAppeles;
      Queue.push v ordreAppel
    else
      begin
        if Queue.is_empty non_boursiers
        then
          let v = Queue.pop boursiers in
          Queue.push v ordreAppel;
          nbBoursiersAppeles := Z.succ !nbBoursiersAppeles;
          nbAppeles := Z.succ !nbAppeles
        else
          begin
            if not resteBoursiers
            then
              begin
                (let o = Queue.pop non_boursiers in Queue.push o ordreAppel);
                nbAppeles := Z.succ !nbAppeles
              end
            else
              begin
                let c = Queue.peek non_boursiers in
                let c' = Queue.peek boursiers in
                let meilleurC =
                  if Z.lt (c.rang) (c'.rang)
                  then Queue.pop non_boursiers
                  else
                    begin
                      nbBoursiersAppeles := Z.succ !nbBoursiersAppeles;
                      Queue.pop boursiers
                    end in
                Queue.push meilleurC ordreAppel;
                nbAppeles := Z.succ !nbAppeles end end end
  done;
  ordreAppel

type groupe_classement = {
  voeux: voeu array;
  taux_boursiers: Z.t;
  taux_residents: Z.t;
  }

let queue_of_array : type a. (a array) ->  (a Queue.t) =
  fun a -> let q = Queue.create () in
           (let o = Z.sub (Z.of_int (Array.length a)) Z.one in
            let o1 = Z.zero in
            let rec for_loop_to i =
              if Z.leq i o
              then begin
                Queue.push (a.(Z.to_int i)) q;
                for_loop_to (Z.succ i)
              end
            in for_loop_to o1);
           q

let diviser_groupe (g: voeu Queue.t) (pred: voeu -> (bool)) :
  (voeu Queue.t) * (voeu Queue.t) =
  let g1 = Queue.create () in
  let g2 = Queue.create () in
  let n = Z.of_int (Queue.length g) in
  (let o = Z.sub n Z.one in let o1 = Z.zero in
   let rec for_loop_to1 i1 =
     if Z.leq i1 o
     then begin
       (let v = Queue.pop g in
        if pred v then Queue.push v g1 else Queue.push v g2);
       for_loop_to1 (Z.succ i1)
     end
   in for_loop_to1 o1);
  (g1, g2)

let est_boursier (p: voeu) : bool = p.boursier

let algo1 (g: groupe_classement) : voeu Queue.t =
  let voeux1 = queue_of_array (g.voeux) in
  let (boursiers,
  non_boursiers) =
  diviser_groupe voeux1 est_boursier in
  ordre_appel boursiers non_boursiers (g.taux_boursiers)

